package polyglot.ext.atomjava.frontend;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.main.*;
import polyglot.types.*;
import polyglot.util.*;

import polyglot.ext.atomjava.frontend.*;
import polyglot.ext.atomjava.types.*;
import polyglot.ext.atomjava.visit.*;

import java.util.*;

public class NotInAtomicOptimizer extends AbstractPass {

  public enum Stage { FIND, ITERATE, TRANSFORM, FINISHED }
  
  protected Stage stage;

  protected Job job;

  public NotInAtomicOptimizer(Pass.ID id, Job job) {
    super(id);

    this.job = job;
    this.stage = Stage.FIND;  

    // prepare the worklists, etc.
    initialize();
  }

  public enum Context { ATOMIC, NOTATOMIC }

  static class ProcedureAndContext {
    ProcedureInstance pi;
    Context c;
    ProcedureAndContext(ProcedureInstance pi, Context c) {
      this.pi = pi;
      this.c = c;
    }
  }

  protected List<ProcedureAndContext> worklist;
  protected List<ProcedureAndContext> nextworklist; 
  protected List<ProcedureAndContext> workedlist;

  // list of jobs todo and finished (per stage/iteration)
  protected List<SourceJob> jobs;

  public void initialize() {

    worklist = new LinkedList<ProcedureAndContext>();
    nextworklist = new LinkedList<ProcedureAndContext>();
    workedlist = new LinkedList<ProcedureAndContext>();

    jobs = new LinkedList<SourceJob>();

    fields_read_inatomic = new HashSet();
    fields_written_inatomic = new HashSet();
    fields_read = new HashSet();
    fields_written = new HashSet();

    fields_arrays_subscript_accessed_inatomic = new HashSet();
    fields_arrays_subscript_assigned_inatomic = new HashSet();

    fields_arrays_written = new HashSet();
    fields_arrays_read = new HashSet();

    // add all non-final java.lang.Object method signatures 
    // and the main method signature
    
    TypeSystem ts;
    MethodInstance mi;

    ts = job.extensionInfo().typeSystem();

    // main method
    mi = ts.methodInstance(Position.COMPILER_GENERATED,
                           ts.Object(),
                           Flags.NONE.Public().Static(),
                           ts.Void(),
                           "main",
                           Collections.singletonList(ts.arrayOf(ts.String())),
                           Collections.EMPTY_LIST);
    nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));

    // hashCode
    mi = ts.methodInstance(Position.COMPILER_GENERATED,
                           ts.Object(),
                           Flags.PUBLIC,
                           ts.Int(),
                           "hashCode",
                           Collections.EMPTY_LIST,
                           Collections.EMPTY_LIST);
    nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));

    // equals
    mi = ts.methodInstance(Position.COMPILER_GENERATED,
                           ts.Object(),
                           Flags.PUBLIC,
                           ts.Boolean(),
                           "equals",
                           Collections.singletonList(ts.Object()),
                           Collections.EMPTY_LIST);
    nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));
    
    // clone
    try {
      Type t = ts.typeForName("java.lang.CloneNotSupportedException");
      mi = ts.methodInstance(Position.COMPILER_GENERATED,
                             ts.Object(),
                             Flags.PROTECTED,
                             ts.Object(),
                             "clone",
                             Collections.EMPTY_LIST,
                             Collections.singletonList(t));
      nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));
    } catch (SemanticException e) {
      throw new InternalCompilerError("Cannot find \"" +
                                      "CloneNotSupportedException\";" +
                                      e.getMessage(), e);
    }
    
    // toString
    mi = ts.methodInstance(Position.COMPILER_GENERATED,
                           ts.Object(),
                           Flags.PUBLIC,
                           ts.String(),
                           "toString",
                           Collections.EMPTY_LIST,
                           Collections.EMPTY_LIST);
    nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));

    // finalize
    mi = ts.methodInstance(Position.COMPILER_GENERATED,
                           ts.Object(),
                           Flags.PROTECTED,
                           ts.Void(),
                           "finalize",
                           Collections.EMPTY_LIST,
                           Collections.singletonList(ts.Throwable()));
    nextworklist.add(new ProcedureAndContext(mi, Context.NOTATOMIC));
  }

  public boolean run() {

    System.out.println("Optimizer Sources:");
    for(Iterator<SourceJob> i = jobs.iterator(); i.hasNext();)
      System.out.println(i.next());

    System.out.println("-----------------------------------------");
    System.out.println();
    System.out.println();

    TypeSystem ts = job.extensionInfo().typeSystem();
    NodeFactory nf = job.extensionInfo().nodeFactory();

    // create our visitors
    FindVisitor finder = new FindVisitor(job, ts, nf);
    IterateVisitor iterator = new IterateVisitor(job, ts, nf);
    TransformVisitor transformer = new TransformVisitor(job, ts, nf);

    while(stage != Stage.FINISHED) {

      switch(stage) {
      case FIND:
        for(Iterator<SourceJob> i = jobs.iterator(); i.hasNext();) {
          SourceJob job = i.next();
          Node ast = job.ast();
          NodeVisitor v_ = finder.begin();
          ast = ast.visit(v_);
          v_.finish(ast);
          job.ast(ast);
        }
        stage = Stage.ITERATE;
        break;
      case ITERATE:

        if(worklist.size() == 0) {
          stage = Stage.TRANSFORM;
          break;
        }

        for(Iterator<SourceJob> i = jobs.iterator(); i.hasNext();) {
          SourceJob job = i.next();
          Node ast = job.ast();
          NodeVisitor v_ = iterator.begin();
          ast = ast.visit(v_);
          v_.finish(ast);
          job.ast(ast);
        }
        break;
      case TRANSFORM:
        for(Iterator<SourceJob> i = jobs.iterator(); i.hasNext();) {
          SourceJob job = i.next();
          Node ast = job.ast();
          NodeVisitor v_ = transformer.begin();
          ast = ast.visit(v_);
          v_.finish(ast);
          job.ast(ast);
        }
        stage = Stage.FINISHED;
        break;
      case FINISHED:
        break;
      default:
        throw new AssertionError("Unknown stage: " + stage);
      }
      
      System.out.println();
      System.out.println("Next stage " + stage + " ; updating the worklist(s)");
      System.out.println();

      workedlist.addAll(worklist);

      worklist.clear();

      next: for(int i = 0; i < nextworklist.size(); /* remove updates i */) {
        
        ProcedureAndContext next = nextworklist.get(i);

        // don't add the procedure and context if it has already been worked
        for(int j = 0; j < workedlist.size(); j++) {
          ProcedureAndContext worked = workedlist.get(j);
          if(ts.equals(worked.pi, next.pi) && worked.c == next.c) {
            nextworklist.remove(i);
            continue next;
          }
        }

        // don't add the procedure (yet) if it exists with a different context
        for(int j = 0; j < worklist.size(); j++) {
          ProcedureAndContext towork = worklist.get(j);
          if(ts.equals(towork.pi, next.pi)) {
            i++;
            continue next;
          }
        }

        nextworklist.remove(i);
        worklist.add(next);
      }

      System.out.println("on worklist:");
      for(Iterator<ProcedureAndContext> i = worklist.iterator(); 
          i.hasNext();) {
        ProcedureAndContext pac = i.next();
        System.out.println("ProcedureInstance: " + pac.pi + " in " + pac.c);
      }

      System.out.println();

      System.out.println("still on nextworklist:");
      for(Iterator<ProcedureAndContext> i = nextworklist.iterator(); 
          i.hasNext();) {
        ProcedureAndContext pac = i.next();
        System.out.println("ProcedureInstance: " + pac.pi + " in " + pac.c);
      }
    }
    
    return true;
  }

  public void addSourceJob(SourceJob job) {
    jobs.add(job);
  }

  //
  // fields_(read|written) are for all fields read or written
  // including in FieldDecl's (to account for constructor on objects)
  //
  protected static Set fields_read_inatomic;
  protected static Set fields_written_inatomic;
  protected static Set fields_read;
  protected static Set fields_written;

  //
  // fields_arrays_subscript_(accessed|assigned)_inatomic gives us the
  // collection of fields which are of an array type that are accessed
  // and assigned via subscript expressions in an atomic context
  //
  protected static Set fields_arrays_subscript_accessed_inatomic;
  protected static Set fields_arrays_subscript_assigned_inatomic;

  //
  // fields_arrays_(written|read) are only for Field expressions 
  // and therefore don't include FieldDecl and therefore tells 
  // whether any writes occur other then from the initialization 
  // expression in the declaration (used to deterimine aliasing)
  //
  protected static Set fields_arrays_written;
  protected static Set fields_arrays_read;

  public static boolean fieldArrayRead(FieldInstance fi) {
    return fields_arrays_read.contains(fi);
  }

  public static boolean fieldArrayWritten(FieldInstance fi) {
    return fields_arrays_written.contains(fi);
  }

  public static boolean fieldArraySubscriptAccessedInAtomic(FieldInstance fi) {
    return fields_arrays_subscript_accessed_inatomic.contains(fi);
  }

  public static boolean fieldArraySubscriptAssignedInAtomic(FieldInstance fi) {
    return fields_arrays_subscript_assigned_inatomic.contains(fi);
  }

  public static boolean fieldReadNotInAtomic(FieldInstance fi) {
    return fields_read.contains(fi);
  }

  public static boolean fieldReadInAtomic(FieldInstance fi) {
    return fields_read_inatomic.contains(fi);
  }

  public static boolean fieldWrittenNotInAtomic(FieldInstance fi) {
    return fields_written.contains(fi);
  }

  public static boolean fieldWrittenInAtomic(FieldInstance fi) {
    return fields_written_inatomic.contains(fi);
  }

  public class FindVisitor extends AJContextVisitor {
    public FindVisitor(Job job, TypeSystem ts, NodeFactory nf) {
      super(job, ts, nf);
    }

    public Node leave(Node parent, Node old, Node n, NodeVisitor v) {
      if(n instanceof MethodDecl) {
        MethodDecl md = (MethodDecl) n;
        // if this method has no implementation then we aren't at fear of
        // calling it. of course, if it extends an untouchable which
        // originally declared the abstract method and a subclass
        // of it implements the method then we will add the untouchable's type
        // and method signature to the worklist (as done below).
        // N.B. by ignoring abstract methods here we also ignore interfaces
        if(md.flags().isAbstract())  // no body == no continue
          return super.leave(parent, old, n, v);
        MethodInstance mi = md.methodInstance();
        if(!ts.isSubtype(mi.container(), typeSystem().AObject())) {
          
          ReferenceType rt = mi.container();
          List methods = rt.methods(mi.name(), mi.formalTypes());
          // the last reference type to have declared/defined this method
          ReferenceType lastrt = rt;
          MethodInstance lastmi = mi;
          while(rt != null) {
            iterating: for(Iterator i = methods.iterator(); i.hasNext();) {
              MethodInstance supermi = (MethodInstance) i.next();
              try {
                if(mi.canOverrideImpl(supermi, true)) {
                  lastrt = rt;
                  lastmi = supermi;
                  break iterating;
                }
              } catch(SemanticException e) {}
            }
            
            if(rt.superType() != null && rt.superType().isReference()) {
              rt = (ReferenceType) rt.superType();
              methods = rt.methods(lastmi.name(), lastmi.formalTypes());
            } else {
              rt = null;
            }
          }
          
          System.out.println("Adding method instance from " + mi.container() + 
                             " last reference type to " + mi + 
                             " was from type " + lastrt + 
                             " with instance " + lastmi);
          nextworklist.add(new ProcedureAndContext(lastmi, Context.NOTATOMIC));
        }
      }      
      return super.leave(parent, old, n, v);
    }

  }

  public class IterateVisitor extends AJContextVisitor {

    public IterateVisitor(Job job, TypeSystem ts, NodeFactory nf) {
      super(job, ts, nf);
    }

    private Context methodOnWorklist(MethodDecl md) {
      MethodInstance mi = md.methodInstance();

      for(Iterator<ProcedureAndContext> i = worklist.iterator(); 
          i.hasNext();) {

        ProcedureAndContext pac = i.next();

        if(pac.pi instanceof MethodInstance &&
           ts.isSubtype(mi.container(), pac.pi.container()) &&
           ts.canOverride(mi, (MethodInstance)pac.pi)) {

          System.out.println("******************************");
          System.out.println("MethodDecl on worklist: " + 
                             mi.container() + " " + mi);
          System.out.println("------------------------------");
          
          return pac.c;
        }
      }
      return null;
    }

    private Context constructorOnWorklist(ConstructorDecl cd) {
      ConstructorInstance ci = cd.constructorInstance();

      for(Iterator<ProcedureAndContext> i = worklist.iterator(); 
          i.hasNext();) {

        ProcedureAndContext pac = i.next();

        if(pac.pi instanceof ConstructorInstance &&
           ts.equals(ci, pac.pi)) {

          System.out.println("******************************");
          System.out.println("ConstructorDecl on worklist: " + 
                             ci.container() + " " + ci);
          System.out.println("------------------------------");

          return pac.c;
        }
      }
      return null;
    }

    private Context constructorOnWorklist(Type container) {

      for(Iterator<ProcedureAndContext> i = worklist.iterator(); 
          i.hasNext();) {
        
        ProcedureAndContext pac = i.next();

        if(pac.pi instanceof ConstructorInstance &&
           ts.equals(container, pac.pi.container()))
          
          return pac.c;
      }

      return null;
    }

    public NodeVisitor enter(Node parent, Node n) {

      ContextVisitor visitor = (ContextVisitor) super.enter(parent, n);

      if(n instanceof MethodDecl) {

        MethodDecl md = (MethodDecl) n;

        Context context = methodOnWorklist(md);
        
        if(context == null)
          return visitor.bypassChildren(n);

        if(context == Context.ATOMIC)
          return visitor.context(context().pushAtomic());

      } else if(n instanceof ConstructorDecl) {
        
        ConstructorDecl cd = (ConstructorDecl) n;

        Context context = constructorOnWorklist(cd);
        
        if(context == null)
          return visitor.bypassChildren(n);

        if(context == Context.ATOMIC)
          return visitor.context(context().pushAtomic());

      } else if(n instanceof FieldDecl) {

        FieldDecl fd = (FieldDecl) n;
        FieldInstance fi = fd.fieldInstance();

        if(fd.initializerInstance() == null)
          return visitor.bypassChildren(n);
        
        Context context = constructorOnWorklist(fi.container());

        if(context == null)
          return visitor.bypassChildren(n);

        if(context == Context.ATOMIC) {
          fields_written_inatomic.add(fd.fieldInstance());
          return visitor.context(context().pushAtomic());
        } else {
          fields_written.add(fd.fieldInstance());
          return visitor;
        }

      } else if(n instanceof Initializer) {
        
        Initializer init = (Initializer) n;
        InitializerInstance ii = init.initializerInstance();
                
        if(init.flags().isStatic())
          return visitor.bypassChildren(n);

        Context context = constructorOnWorklist(ii.container());

        if(context == null)
          return visitor.bypassChildren(n);

        if(context == Context.ATOMIC)
          return visitor.context(context().pushAtomic());

      } else if(n instanceof ArrayAccess) {
        ArrayAccess array_access = (ArrayAccess) n;

        if(array_access.array() instanceof Field) {
          
          Field f = (Field) array_access.array();

          if(f.isTargetImplicit()) {

            if(context().inAtomicContext())
              fields_arrays_subscript_accessed_inatomic.add(f.fieldInstance());

            // 1) update the fields_read(_inatomic) list
            //    don't update the fields_arrays_read

            if(context().inAtomicContext())
              fields_read_inatomic.add(f.fieldInstance());
            else
              fields_read.add(f.fieldInstance());

            // 2) bypass the visitor

            return visitor.bypass(f);
          }
        }

      } else if(n instanceof ArrayAccessAssign) {
        ArrayAccessAssign array_access_assign = (ArrayAccessAssign) n;

        if(((ArrayAccess)array_access_assign.left()).array() instanceof Field) {
          Field f = (Field) ((ArrayAccess)array_access_assign.left()).array();

          if(f.isTargetImplicit()) {

            // update subscript assigned if in atomic

            if(context().inAtomicContext())
              fields_arrays_subscript_assigned_inatomic.add(f.fieldInstance());

            // 1) update the fields_read(_inatomic) list
            //    don't update the fields_arrays_read

            if(context().inAtomicContext())
              fields_read_inatomic.add(f.fieldInstance());
            else
              fields_read.add(f.fieldInstance());

            // 2) bypass the visitor

            return visitor.bypass(f);            
          }
        }

      } else if(n instanceof CodeDecl) {
        throw new InternalCompilerError("Unknown CodeDecl " + n);
      }

      return visitor;
    }
    
    public Node leave(Node parent, Node old, Node n, NodeVisitor v) {

      if(n instanceof Call) {
        Call c = (Call) n;
        System.out.println("Call ; Adding method instance " + 
                           c.methodInstance().container() + " " +
                           c.methodInstance() + 
                           " to the worklist.");
        
        nextworklist.add(new ProcedureAndContext(c.methodInstance(),
                                                 context().inAtomicContext() ? 
                                                 Context.ATOMIC : 
                                                 Context.NOTATOMIC));
      } else if(n instanceof Field) {
        Field f = (Field) n;            

        if(parent instanceof Assign && f == ((Assign)parent).left()) {
          if(context().inAtomicContext())
            fields_written_inatomic.add(f.fieldInstance());
          else
            fields_written.add(f.fieldInstance());

          if(f.fieldInstance().type() instanceof ArrayType)
            fields_arrays_written.add(f.fieldInstance());
        } else {
          if(context().inAtomicContext())
            fields_read_inatomic.add(f.fieldInstance());
          else
            fields_read.add(f.fieldInstance());

          if(f.fieldInstance().type() instanceof ArrayType)
            fields_arrays_read.add(f.fieldInstance());
        }


      } else if(n instanceof New) {
        New ne = (New) n;
        System.out.println("New ; Adding constructor instance " + 
                           ne.constructorInstance().container() + " " +
                           ne.constructorInstance() + 
                           " to the worklist.");
        nextworklist.add(new ProcedureAndContext(ne.constructorInstance(),
                                                 context().inAtomicContext() ? 
                                                 Context.ATOMIC : 
                                                 Context.NOTATOMIC));
      } else if(n instanceof ConstructorCall) {
        ConstructorCall cc = (ConstructorCall) n;
        System.out.println("super(...) or this(...) ; " + 
                           "Adding constructor instance " + 
                           cc.constructorInstance().container() + " " +
                           cc.constructorInstance() + 
                           " to the worklist.");

        nextworklist.add(new ProcedureAndContext(cc.constructorInstance(),
                                                 context().inAtomicContext() ? 
                                                 Context.ATOMIC : 
                                                 Context.NOTATOMIC));
      }
      
      return super.leave(parent, old, n, v);
    }
  }

  public class TransformVisitor extends AJContextVisitor {
    public TransformVisitor(Job job, TypeSystem ts, NodeFactory nf) {
      super(job, ts, nf);
    }

    public void finish() {
      System.out.println();
      System.out.println();

      System.out.println("--------- READ IN-ATOMIC ---------");
      for(Iterator i = fields_read_inatomic.iterator(); i.hasNext();) {
        FieldInstance fi = (FieldInstance) i.next();
        System.out.println(fi.container() + " " + fi);
      }

      System.out.println();
      System.out.println("--------- WRITTEN IN-ATOMIC ---------");
      for(Iterator i = fields_written_inatomic.iterator(); i.hasNext();) {
        FieldInstance fi = (FieldInstance) i.next();
        System.out.println(fi.container() + " " + fi);
      }

      System.out.println();
      System.out.println("--------- READ ---------");
      for(Iterator i = fields_read.iterator(); i.hasNext();) {
        FieldInstance fi = (FieldInstance) i.next();
        System.out.println(fi.container() + " " + fi);
      }

      System.out.println();
      System.out.println("--------- WRITTEN ---------");
      for(Iterator i = fields_written.iterator(); i.hasNext();) {
        FieldInstance fi = (FieldInstance) i.next();
        System.out.println(fi.container() + " " + fi);
      }

    }
  }
}
