package polyglot.ext.atomjava.types;

import polyglot.types.*;

import polyglot.frontend.Compiler;
import polyglot.frontend.ExtensionInfo;
import polyglot.frontend.FileSource;
import polyglot.main.Report;
import polyglot.types.reflect.ClassFile;
import polyglot.types.reflect.ClassFileLoader;
import polyglot.util.*;

public class AJSourceClassResolver extends SourceClassResolver {


  public AJSourceClassResolver(Compiler compiler, ExtensionInfo ext,
                               String classpath, ClassFileLoader loader,
                               boolean allowRawClasses)
  {
    super(compiler, ext, classpath, loader, allowRawClasses);
  }

  /**
   * Find a type by name.
   */
  public Named find(String name) throws SemanticException {
    if (Report.should_report(report_topics, 3))
	Report.report(3, "SourceCR.find(" + name + ")");

    ClassFile clazz = null;
    ClassFile encodedClazz = null;
    FileSource source = null;

    // First try the class file.
    clazz = loadFile(name);
    if (clazz != null) {
      // Check for encoded type information.
      if (clazz.encodedClassType(version.name()) != null) {
        if (Report.should_report(report_topics, 4))
          Report.report(4, "Class " + name + " has encoded type info");
        encodedClazz = clazz;
      }
    }
    
    if (clazz != null && ! this.allowRawClasses) {
      clazz = null;
    }

    polyglot.ext.atomjava.main.AJOptions options = 
      (polyglot.ext.atomjava.main.AJOptions) 
      ((polyglot.ext.atomjava.ExtensionInfo)ext).getOptions();
    if(options.optimize) {
      if(encodedClazz != null) {
        // for optimization purposes:
	// if polyglot has compiled it then compile it again
	// we force this because while a user supplied classpath
	// can not be included (and therefore even if they did have
	// some encoded class lying around we won't get it), we might
	// have compiled something in the runtime/j2se that has been
	// atomized and we want to recompile it for THIS optimization
        encodedClazz = null;
        clazz = null;
      }
    }

    // Now, try and find the source file.
    source = ext.sourceLoader().classSource(name);
    if (Report.should_report(report_topics, 4)) {
        if (source == null) 
            Report.report(4, "Class " + name + " not found in source file");
        else
            Report.report(4, "Class " + name + " found in source " + source);
    }

    // Don't use the raw class if the encoded class is available.
    if (encodedClazz != null) {
      if (Report.should_report(report_topics, 4))
	Report.report(4, "Not using raw class file for " + name);
      clazz = null;
    }

    // If both the source and class are available, we decide which to
    // use based on compiler compatibility and modification times.
    if (clazz != null && source != null) {
      long classModTime = clazz.rawSourceLastModified();
      long sourceModTime = source.lastModified().getTime();

      if (classModTime < sourceModTime) {
        if (Report.should_report(report_topics, 3))
	    Report.report(3, "Source file version is newer than compiled for " +
                      name + ".");
        clazz = null;
      }
      else {
        source = null;
      }
    }

    if (encodedClazz != null && source != null) {
      long classModTime = encodedClazz.sourceLastModified(version.name());
      long sourceModTime = source.lastModified().getTime();

      int comp = checkCompilerVersion(encodedClazz.compilerVersion(version.name()));

      if (classModTime < sourceModTime) {
        if (Report.should_report(report_topics, 3))
	    Report.report(3, "Source file version is newer than compiled for " +
                      name + ".");
        encodedClazz = null;
      }
      else if (comp != COMPATIBLE) {
        // Incompatible or older version, so go with the source.
        if (Report.should_report(report_topics, 3))
	    Report.report(3, "Incompatible source file version for " + name + ".");
        encodedClazz = null;
      }
    }

    if (encodedClazz != null) {
      if (Report.should_report(report_topics, 4))
	Report.report(4, "Using encoded class type for " + name);
      try {
        return getEncodedType(encodedClazz, name);
      }
      catch (BadSerializationException e) {
      	throw e;
      }
      catch (SemanticException e) {
        if (Report.should_report(report_topics, 4))
          Report.report(4, "Could not load encoded class " + name);
        encodedClazz = null;
      }
    }

    // At this point, at most one of clazz and source should be set.

    if (clazz != null) {
      if (Report.should_report(report_topics, 4))
	Report.report(4, "Using raw class file for " + name);
      return clazz.type(ts);
    }

    if (source != null) {
      if (Report.should_report(report_topics, 4))
	Report.report(4, "Using source file for " + name);
      return getTypeFromSource(source, name);
    }

    if (clazz != null && !this.allowRawClasses) {
        // We have a raw class only. We do not have the source code,
        // or encoded class information. 
        throw new SemanticException("Class \"" + name + "\" not found."
            + " A class file was found, but it did not contain appropriate" 
            + " information for the Polyglot-based compiler " + 
            ext.compilerName() + ". Try using " + ext.compilerName() 
            + " to recompile the source code.");
    }
    throw new NoClassException(name);
  }
}
